mod bot;
mod four;
mod game;
mod pl2;

use bot::Bottable;
use crossterm::tty::IsTty;
use four::{Four, FourMove};
use game::*;
use pl2::BinaryPlayer;
use std::io::{self, BufRead, Write};

#[derive(Debug)]
enum Miscommunication {
    Io(io::Error),
    MoveParse,
    ModeParse,
}

impl From<io::Error> for Miscommunication {
    fn from(error: io::Error) -> Self {
        Self::Io(error)
    }
}

fn handle_outcome<P: GamePlayer>(mmeee: P, out: MoveOutcome<P>) -> (&'static str, bool) {
    match out {
        MoveOutcome::Win(p) => (
            if p == mmeee {
                "YIPPIEE I WINNER"
            } else {
                "nuuuu :("
            },
            false,
        ),
        MoveOutcome::Tie => ("meh", false),
        MoveOutcome::Pass => ("...", true),
    }
}

fn apply_debugged<P: GamePlayer, M: GameMove, T: Game<P, M> + std::fmt::Display>(
    mmeee: P,
    game: &mut T,
    m: M,
    istty: bool,
) -> bool {
    let (msg, out) = handle_outcome(mmeee, game.apply(m).expect("unexpected move failure"));
    if istty {
        eprintln!("{}{}", &game, msg);
    }
    out
}

use clap::Parser;

#[derive(Parser, Debug)]
#[command(author, version, about, long_about = None)]
struct Args {
    #[arg(short, long)]
    quite: bool,
    #[arg(short, long)]
    load: bool,
}

fn main() -> Result<(), Miscommunication> {
    let args = Args::parse();
    assert!(
        !(args.quite && args.load),
        "can't be load and quite at once..."
    );
    let mut game: Four<BinaryPlayer> = Four::new();
    let reader = io::stdin();
    let mut stdout = io::stdout();
    let istty = if !(args.quite || args.load) {
        reader.is_tty()
    } else {
        args.load
    };
    let mut reader = reader.lock().lines();
    let mmeee;
    loop {
        if istty {
            eprint!("> ");
        }
        match reader.next().ok_or(Miscommunication::ModeParse)??.as_str() {
            "f" => {
                mmeee = BinaryPlayer::first();
                let m = game.best_move().expect("first move fail");
                println!("{}", m);
                stdout.flush()?;
                assert!(
                    apply_debugged(mmeee, &mut game, m, istty),
                    "first move fail"
                );
                break;
            }
            "s" => {
                mmeee = BinaryPlayer::first().next();
                if istty {
                    eprintln!("{}", &game);
                }
                break;
            }
            _ => {
                if istty {
                    eprintln!("bad mode");
                }
            }
        }
    }
    let mut reader = reader.map(|l| l.map(|s| s.parse::<u8>()));
    loop {
        loop {
            let m = loop {
                if istty {
                    eprint!("{}> ", game.current_player());
                }
                let num = reader.next().ok_or(Miscommunication::MoveParse)??.ok();
                if let Some(m) = num.map(|num| FourMove::try_from(num).ok()).flatten() {
                    break m;
                }
            };
            let out = game.apply(m);
            match out {
                Some(out) => {
                    let (msg, out) = handle_outcome(mmeee, out);
                    if !out {
                        if istty {
                            eprintln!("{}{}", &game, msg);
                        }
                        return Ok(());
                    }
                    break;
                }
                None => (),
            }
        }
        let m = game.best_move().expect("no moves");
        println!("{}", m);
        stdout.flush()?;
        if !apply_debugged(mmeee, &mut game, m, istty) {
            break;
        }
    }
    Ok(())
}
